#!/usr/bin/env python
# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d.axes3d as p3  # noqa: F401
import numpy as np
from matplotlib import animation

from attractors.utils.colortable import get_continuous_cmap
from attractors.utils.des import DES
from attractors.utils.video import ffmpeg_video


def animate_simulation(
    attractor,
    init_coord,
    attr_params,
    xlim,
    ylim,
    zlim,
    width,
    height,
    dpi,
    bgcolor,
    palette,
    sim_time,
    points,
    n,
    des,
    live,
    rk2_method,
    fps,
    outf,
):

    fig = plt.figure(figsize=(width, height), dpi=dpi)
    ax = fig.add_axes([0, 0, 1, 1], projection="3d")
    ax.axis("off")
    fig.set_facecolor(bgcolor)
    ax.set_facecolor(bgcolor)

    init_coords = [init_coord] + [
        init_coord + np.random.normal(0, 0.01, 3) for _ in range(n - 1)
    ]

    attractor_vects = [DES(xyz, attractor, attr_params) for xyz in init_coords]

    for vect in attractor_vects:
        try:
            rk = getattr(vect, des)
            if des == "rk2":
                rk(0, sim_time, points, rk2_method)
            else:
                rk(0, sim_time, points)
        except AttributeError as e:
            raise Exception(
                f"Integrator Error. {des} is not an valid integrator"
            ) from e

    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    ax.set_zlim(zlim)

    if isinstance(palette, str):
        cmap = plt.cm.get_cmap(palette)
    else:
        cmap = get_continuous_cmap(palette)

    colors = cmap(np.linspace(0, 1, len(init_coords)))

    lines = sum(
        [ax.plot([], [], [], "-", c=c, linewidth=1, antialiased=True) for c in colors],
        [],
    )
    pts = sum([ax.plot([], [], [], "o", c=c) for c in colors], [])

    trail = int(0.9 * points)

    def init():
        for line, pt in zip(lines, pts):
            line.set_data_3d([], [], [])
            pt.set_data_3d([], [], [])
        return lines + pts

    def update(i):
        i = i % len(attractor_vects[0].X)
        for line, pt, k in zip(lines, pts, attractor_vects):
            if i > trail:
                line.set_data_3d(
                    k.X[i - trail : i],  # noqa: E203
                    k.Y[i - trail : i],  # noqa: E203
                    k.Z[i - trail : i],  # noqa: E203
                )
            else:
                line.set_data_3d(k.X[:i], k.Y[:i], k.Z[:i])
            pt.set_data_3d(k.X[i], k.Y[i], k.Z[i])
        ax.view_init(0.005 * i, 0.05 * i)
        return lines + pts

    if live:
        _ = animation.FuncAnimation(
            fig, update, init_func=init, interval=1000 / fps, blit=False
        )
        plt.show()
    else:
        ffmpeg_video(fig, update, points, fps, outf)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import json

import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d.axes3d as p3  # noqa: F401
import numpy as np
from matplotlib import animation
from mpl_toolkits.mplot3d.art3d import Line3DCollection

from attractors import data
from attractors.utils.base import ATTRACTOR_PARAMS
from attractors.utils.colortable import get_continuous_cmap
from attractors.utils.des import DES
from attractors.utils.video import ffmpeg_video

try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources

# * load theme
raw_themes_data = pkg_resources.open_text(data, "themes.json")
themes = json.load(raw_themes_data)


class Attractor(DES):

    bgcolor = None
    palette = None
    fig, ax = None, None

    def __init__(self, initial_coord, attractor, **kwargs):
        self.attr = ATTRACTOR_PARAMS[attractor]
        params = {
            self.attr["params"][i]: kwargs.get(
                self.attr["params"][i], self.attr["default_params"][i]
            )
            for i in range(len(self.attr["params"]))
        }
        super(Attractor, self).__init__(initial_coord, attractor, params)

    def __eq__(self, other):
        if not isinstance(other, Attractor):
            return NotImplemented
        return self.attractor == other.attractor

    @classmethod
    def set_theme(cls, theme, bgcolor, palette):
        if all(v is None for v in [theme, bgcolor, palette]):
            cls.bgcolor = "#000000"
            cls.palette = "jet"
        elif all(v is None for v in [bgcolor, palette]) and theme is not None:
            palette_temp = list(theme.values())
            palette_temp.remove(theme["background"])
            cls.bgcolor = theme["background"]
            cls.palette = palette_temp
        else:
            if bgcolor is not None:
                cls.bgcolor = bgcolor
            if palette is not None:
                cls.palette = palette

    @classmethod
    def set_figure(cls, width, height, dpi):
        cls.fig = plt.figure(figsize=(width, height), dpi=dpi)
        cls.ax = cls.fig.add_axes([0, 0, 1, 1], projection="3d")
        cls.ax.axis("off")

        cls.fig.set_facecolor(cls.bgcolor)
        cls.ax.set_facecolor(cls.bgcolor)

        if isinstance(cls.palette, str):
            cls.cmap = plt.cm.get_cmap(cls.palette)
        else:
            cls.cmap = get_continuous_cmap(cls.palette)

    @classmethod
    def set_limits(cls, xlim, ylim, zlim):
        cls.ax.set_xlim(xlim)
        cls.ax.set_ylim(ylim)
        cls.ax.set_zlim(zlim)

    @classmethod
    def _wrap_set(cls, objs, kwargs):

        assert all(
            x == objs[0] for x in objs
        ), "All objects must be of the same attractor type"

        try:
            theme = themes[kwargs.get("theme")]
        except KeyError:
            theme = None

        Attractor.set_theme(
            theme,
            bgcolor=kwargs.get("bgcolor", None),
            palette=kwargs.get("palette", None),
        )
        Attractor.set_figure(
            width=kwargs.get("width", 16),
            height=kwargs.get("width", 9),
            dpi=kwargs.get("dpi", 120),
        )

        attr = ATTRACTOR_PARAMS[objs[0].attractor]
        Attractor.set_limits(
            xlim=kwargs.get("xlim", attr["xlim"]),
            ylim=kwargs.get("ylim", attr["ylim"]),
            zlim=kwargs.get("zlim", attr["zlim"]),
        )

    @classmethod
    def set_animate_multipoint(cls, *objs, **kwargs):

        Attractor._wrap_set(objs, kwargs)
        colors = cls.cmap(np.linspace(0, 1, len(objs)))

        lines = sum(
            [
                cls.ax.plot([], [], [], "-", c=c, linewidth=1, antialiased=True)
                for c in colors
            ],
            [],
        )
        pts = sum([cls.ax.plot([], [], [], "o", c=c) for c in colors], [])

        def init():
            for line, pt in zip(lines, pts):
                line.set_data_3d([], [], [])
                pt.set_data_3d([], [], [])
            return lines + pts

        def update(i):
            i = i % len(objs[0].X)
            for line, pt, k in zip(lines, pts, objs):
                line.set_data_3d(k.X[:i], k.Y[:i], k.Z[:i])
                pt.set_data_3d(k.X[i], k.Y[i], k.Z[i])
            cls.ax.view_init(0.005 * i, 0.05 * i)
            return lines + pts

        points = len(max(objs).X)
        return update, points, init

    @classmethod
    def set_animate_gradient(cls, obj, **kwargs):

        Attractor._wrap_set([obj], kwargs)

        line = Line3DCollection([], cmap=cls.cmap)
        cls.ax.add_collection3d(line)

        (pt,) = cls.ax.plot([], [], [], "o")
        line.set_array(np.array(obj.Z))
        colors = line.to_rgba(obj.Z)

        def init():
            line.set_segments([])
            pt.set_data_3d([], [], [])
            return line, pt

        def update(i):
            i = i % len(obj.X)
            pts = (
                np.array([obj.X[:i], obj.Y[:i], obj.Z[:i]])
                .transpose()
                .reshape(-1, 1, 3)
            )
            segs = np.concatenate([pts[:-1], pts[1:]], axis=1)
            line.set_segments(segs)
            pt.set_data_3d([obj.X[i]], [obj.Y[i]], [obj.Z[i]])
            pt.set_color(colors[i])
            cls.ax.view_init(0.005 * i, 0.05 * i)
            return line, pt

        points = len(obj.X)
        return update, points, init

    @classmethod
    def animate(cls, update, points, init=None, **kwargs):
        if kwargs.get("live", False):
            _ = animation.FuncAnimation(
                cls.fig,
                update,
                init_func=init,
                interval=1000 / kwargs.get("fps", 60),
                blit=False,
            )
            plt.show()
        else:
            ffmpeg_video(
                cls.fig,
                update,
                points,
                kwargs.get("fps", 60),
                kwargs.get("outf", "output.mp4"),
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d.axes3d as p3  # noqa: F401
import numpy as np
from mpl_toolkits.mplot3d.art3d import Line3DCollection

from attractors.utils.colortable import get_continuous_cmap
from attractors.utils.des import DES
from attractors.utils.video import ffmpeg_video


def animate_gradient(
    attractor,
    init_coord,
    attr_params,
    xlim,
    ylim,
    zlim,
    width,
    height,
    dpi,
    bgcolor,
    palette,
    sim_time,
    points,
    des,
    rk2_method,
    fps,
    outf,
):
    fig = plt.figure(figsize=(width, height), dpi=dpi)
    ax = fig.add_axes([0, 0, 1, 1], projection="3d")
    ax.axis("off")
    fig.set_facecolor(bgcolor)
    ax.set_facecolor(bgcolor)

    vect = DES(init_coord, attractor, attr_params)
    try:
        rk = getattr(vect, des)
        if des == "rk2":
            rk(0, sim_time, points, rk2_method)
        else:
            rk(0, sim_time, points)
    except AttributeError as e:
        raise Exception(f"Integrator Error. {des} is not an valid integrator") from e

    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    ax.set_zlim(zlim)

    if isinstance(palette, str):
        cmap = plt.cm.get_cmap(palette)
    else:
        cmap = get_continuous_cmap(palette)

    line = Line3DCollection([], cmap=cmap)
    ax.add_collection3d(line)

    (pt,) = ax.plot([], [], [], "o")

    def init():
        line.set_segments([])
        pt.set_data_3d([], [], [])

    init()

    line.set_array(np.array(vect.Y))
    colors = line.to_rgba(vect.Z)

    def update(frame):

        i = frame % len(vect.X)
        pts = (
            np.array([vect.X[:i], vect.Y[:i], vect.Z[:i]]).transpose().reshape(-1, 1, 3)
        )
        segs = np.concatenate([pts[:-1], pts[1:]], axis=1)

        line.set_segments(segs)

        pt.set_data_3d([vect.X[i]], [vect.Y[i]], [vect.Z[i]])
        pt.set_color(colors[i])

        ax.view_init(0.005 * i, 0.05 * i)

    ffmpeg_video(fig, update, points, fps, outf)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import json

import numpy as np

try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources

from attractors import data

raw_params_data = pkg_resources.open_text(data, "params.json")
ATTRACTOR_PARAMS = json.load(raw_params_data)


class BaseAttractors(object):
    def __init__(self, attractor, params):

        self.attractor = attractor
        self._func_params(params)

    def _func_params(self, params):
        try:
            for prm in ATTRACTOR_PARAMS[self.attractor]["params"]:
                exec("self.{} = {}".format(prm, params[prm]))
        except KeyError as e:
            raise Exception(
                "Parameter argument error. Invalid parameter"
                f" for {self.attractor} attractor."
            ) from e

    def lorenz(self, coord):
        x, y, z = coord
        dx = self.sigma * (y - x)
        dy = x * (self.rho - z) - z
        dz = x * y - (self.beta * z)
        return np.array([dx, dy, dz], dtype="double")

    def rabinovich_fabrikant(self, coord):
        x, y, z = coord
        dx = y * (z - 1 + (x * x)) + (self.gamma * x)
        dy = x * (3 * z + 1 - (x * x)) + (self.gamma * y)
        dz = -2 * z * (self.alpha + x * y)
        return np.array([dx, dy, dz], dtype="double")

    def lotka_volterra(self, coord):
        x, y, z = coord
        dx = x - x * y + self.c * x * x - self.a * z * x * x
        dy = -y + x * y
        dz = -self.b * z + self.a * z * x * x
        return np.array([dx, dy, dz], dtype="double")

    def rossler(self, coord):
        x, y, z = coord
        dx = -(y + z)
        dy = x + (self.a * y)
        dz = self.b + z * (x - self.c)
        return np.array([dx, dy, dz], dtype="double")

    def wang_sun(self, coord):
        x, y, z = coord
        dx = self.a * x + self.c * y * z
        dy = self.b * x + self.d * y - x * z
        dz = self.e * z + self.f * x * y
        return np.array([dx, dy, dz], dtype="double")

    def rikitake(self, coord):
        x, y, z = coord
        dx = -self.mu * x + z * y
        dy = -self.mu * y + x * (z - self.a)
        dz = 1 - x * y
        return np.array([dx, dy, dz], dtype="double")

    def nose_hoover(self, coord):
        x, y, z = coord
        dx = self.a * y
        dy = -x + y * z
        dz = 1 - y * y
        return np.array([dx, dy, dz], dtype="double")

    def aizawa(self, coord):
        x, y, z = coord
        dx = (z - self.b) * x - self.d * y
        dy = self.d * x + (z - self.b) * y
        dz = (
            self.c
            + self.a * z
            - (z ** 3 / 3)
            - (x ** 2 + y ** 2) * (1 + self.e * z)
            + self.f * z * x ** 3
        )
        return np.array([dx, dy, dz], dtype="double")

    def three_cell_cnn(self, coord):
        x, y, z = coord
        fx = 0.5 * (np.abs(x + 1) - np.abs(x - 1))
        fy = 0.5 * (np.abs(y + 1) - np.abs(y - 1))
        fz = 0.5 * (np.abs(z + 1) - np.abs(z - 1))
        dx = -x + self.p1 * fx - self.s * fy - self.s * fz
        dy = -y - self.s * fx + self.p2 * fy - self.r * fz
        dz = -z - self.s * fx + self.r * fy + fz
        return np.array([dx, dy, dz], dtype="double")

    def bouali_type_1(self, coord):
        x, y, z = coord
        dx = self.k * y + self.mu * x * (self.b - y * y)
        dy = -x + self.s * z
        dz = self.p * x - self.q * y
        return np.array([dx, dy, dz], dtype="double")

    def bouali_type_2(self, coord):
        x, y, z = coord
        dx = x * (self.a - y) + self.alpha * z
        dy = -y * (self.b - x * x)
        dz = -x * (self.c - self.s * z) - self.beta * z
        return np.array([dx, dy, dz], dtype="double")

    def bouali_type_3(self, coord):
        x, y, z = coord
        dx = self.alpha * x * (1 - y) - self.beta * z
        dy = -self.gamma * y * (1 - x * x)
        dz = self.mu * x
        return np.array([dx, dy, dz], dtype="double")

    def finance(self, coord):
        x, y, z = coord
        dx = (1 / self.b - self.a) * x + x * y + z
        dy = -self.b * y - x * x
        dz = -x - self.c * z
        return np.array([dx, dy, dz], dtype="double")

    def burke_shaw(self, coord):
        x, y, z = coord
        dx = -self.s * (x + y)
        dy = -y - self.s * x * z
        dz = self.s * x * y + self.v
        return np.array([dx, dy, dz], dtype="double")

    def moore_spiegel(self, coord):
        x, y, z = coord
        dx = y
        dy = z
        dz = -z - (self.t - self.r * (1 - x * x)) * y - self.t * x
        return np.array([dx, dy, dz], dtype="double")

    def sakarya(self, coord):
        x, y, z = coord
        dx = -x + y + y * z
        dy = -x - y + self.a * x * z
        dz = z - self.b * x * y
        return np.array([dx, dy, dz], dtype="double")

    def dadras(self, coord):
        x, y, z = coord
        dx = y - self.a * x + self.b * y * z
        dy = self.c * y - x * z + z
        dz = self.d * x * y - self.h * z
        return np.array([dx, dy, dz], dtype="double")

    def halvorsen(self, coord):
        x, y, z = coord
        dx = -self.a * x - 4 * y - 4 * z - y * y
        dy = -self.a * y - 4 * z - 4 * x - z * z
        dz = -self.a * z - 4 * x - 4 * y - x * x
        return np.array([dx, dy, dz], dtype="double")

    def hadley(self, coord):
        x, y, z = coord
        dx = -y * y - z * z - self.a * (x - self.f)
        dy = x * y - self.b * x * z - y + self.g
        dz = self.b * x * y + z * (x - 1)
        return np.array([dx, dy, dz], dtype="double")

    def chen(self, coord):
        x, y, z = coord
        dx = self.a * (y - x)
        dy = (self.c - self.a) * x - (x * z) + self.c * y
        dz = x * y - self.b * z
        return np.array([dx, dy, dz], dtype="double")

    def chen_lee(self, coord):
        x, y, z = coord
        dx = self.a * x - y * z
        dy = self.b * y + x * z
        dz = self.c * z + x * y / 3
        return np.array([dx, dy, dz], dtype="double")

    def chen_celikovsky(self, coord):
        x, y, z = coord
        dx = self.a * (y - x)
        dy = -(x * z) + self.c * y
        dz = x * y - self.b * z
        return np.array([dx, dy, dz], dtype="double")

    def thomas_cyclically_symmetric(self, coord):
        x, y, z = coord
        dx = np.sin(y) - self.b * x
        dy = np.sin(z) - self.b * y
        dz = np.sin(x) - self.b * z
        return np.array([dx, dy, dz], dtype="double")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import subprocess


def ffmpeg_video(fig, update_func, points, fps, outf):
    """Generates output video given a animation function via ffmpeg

    Args:
        fig (matplotlib.figure.Figure): matplotlib figure instance
        update_func (callable): animation function
        points (int): number of points used for the simulation
        fps (int): frames per second for output video
        outf (str): output video filename
    """
    canvas_width, canvas_height = fig.canvas.get_width_height()
    cmdstring = (
        "ffmpeg",
        "-y",
        "-r",
        "%d" % fps,
        "-s",
        "%dx%d" % (canvas_width, canvas_height),
        "-pix_fmt",
        "argb",  # format
        "-f",
        "rawvideo",
        "-i",
        "-",
        "-b:v",
        "5000k",
        "-vcodec",
        "mpeg4",
        outf,
    )
    p = subprocess.Popen(cmdstring, stdin=subprocess.PIPE)

    for frame in range(points):
        update_func(frame)
        fig.canvas.draw()
        string = fig.canvas.tostring_argb()
        p.stdin.write(string)

    p.communicate()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['attractors', 'attractors.anim', 'attractors.data', 'attractors.utils']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.4.2,<4.0.0']

entry_points = \
{'console_scripts': ['attractors = attractors.parser:cli']}

setup_kwargs = {
    'name': 'attractors',
    'version': '1.0.0',
    'description': 'Simulation and visualization of strange attractors',
    'long_description': 'Attractors\n==========\n\nAttractors is a package for simulation and visualization of strange\nattractors.\n\n|PyPI license| |PyPI version| |image1|\n\n.. |PyPI license| image:: https://img.shields.io/github/workflow/status/Vignesh-Desmond/attractors/Build?style=flat-square&logo=GitHub\n   :target: https://github.com/Vignesh-Desmond/attractors/actions/workflows/build.yml\n.. |PyPI version| image:: https://img.shields.io/pypi/v/attractors?color=blue&style=flat-square\n   :target: https://pypi.python.org/pypi/attractors/\n.. |image1| image:: https://img.shields.io/pypi/l/attractors?style=flat-square&color=orange\n   :target: https://lbesson.mit-license.org/\n\nFor complete documentation, see the repository docs at\nhttps://github.com/Vignesh-Desmond/attractors\n\nInstallation\n------------\n\nThe package is currently hosted on PyPi and can be installed with pip:\n\n``pip install attractors``\n\nAlternatively, the package can be installed locally after cloning by the\nfollowing command:\n\n``python -m pip install .``\n\nNote:\n\nThe package requires FFMPEG as a prerequisite for generating\nvisualization output. For installation, see\n`here <https://ffmpeg.org/download.html>`__\n\nDependencies\n------------\n\n-  Python (3.8+)\n-  NumPy (1.21.0+)\n-  Matplotlib (3.4.2+)\n\nUsage\n-----\n\nThe package is intended to be used mainly via the command-line\ninterface. The simplest way to visualize an Lorenz attractor is given\nbelow:\n\n::\n\n   attractors -p 100000 -s 100 -t multipoint lorenz\n\n``-p`` : Number of points for the simulation\n\n``-s`` : Simulation time\n\n``-t`` : Type of visualization\n\nThere are two types of visualizations:\n\n``multipoint`` visualization can simulate multiple initial coordinates,\nuseful to infer the chaotic nature of the attractors. Use ``--n`` to set\nthe number of initial points.\n\n``gradient`` visualization uses gradient color for the 3D plot, given a\nspecific axis.\n\nBy default, the visualization output will be saved in an MPEG4 encoded\nvideo. To show a live plot, use the ``--live`` flag (only supported for\nmultipoint).\n\nThe full list of possible settings can be obtained by the help command:\n``attractors -h``\n\n::\n\n   usage: attractors [-v] [-h] -t {multipoint,gradient}\n                     [--des {euler,rk2,rk3,rk4,rk5}] [--width WIDTH]\n                     [--height HEIGHT] [--dpi DPI] [--theme THEME] -s SIMTIME -p\n                     SIMPOINTS [--bgcolor BGCOLOR] [--cmap CMAP] [--fps FPS]\n                     [--n N] [--rk2 {heun,imp_poly,ralston}] [--outf OUTF]\n                     [--live]\n                     ATTRACTOR ...\n\n   optional arguments:\n     -v, --version         show program\'s version number and exit\n     -h, --help            show this help message and exit\n\n   required arguments:\n     -t {multipoint,gradient}, --type {multipoint,gradient}\n                           choose simulation type\n     -s SIMTIME, --simtime SIMTIME\n                           set the simulation time\n     -p SIMPOINTS, --simpoints SIMPOINTS\n                           set the number of points to be used for the simulation\n\n   other arguments:\n     --des {euler,rk2,rk3,rk4,rk5}\n                           set the Differential Equation Solver. Default: rk4\n     --width WIDTH         set width of the figure Default: 16\n     --height HEIGHT       set height of the figure Default: 9\n     --dpi DPI             set DPI of the figure Default: 120\n     --theme THEME         choose theme (color palette) to be used\n     --bgcolor BGCOLOR     Background color for figure in hex. Overrides theme\n                           settings Default: #000000\n     --cmap CMAP           Matplotlib cmap for palette. Overrides theme settings\n                           Default: jet\n     --fps FPS             Set FPS for animated video (or interactive plot)\n                           Default: 60\n     --n N                 Number of initial points for Multipoint animation\n                           Default: 3\n     --rk2 {heun,imp_poly,ralston}\n                           Method for 2nd order Runge-Kutta if specified to used.\n                           Default: heun\n     --outf OUTF           Output video filename Default: output.mp4\n     --live                Live plotting instead of generating video.\n\n   Attractor settings:\n     Choose one of the attractors and specify its parameters\n\n     ATTRACTOR\n       lorenz              Lorenz attractor\n       rabinovich_fabrikant\n                           Rabinovich Fabrikant attractor\n       lotka_volterra      Lotka Volterra attractor\n       rossler             Rossler attractor\n       wang_sun            Wang Sun attractor\n       rikitake            Rikitake attractor\n       nose_hoover         Nose Hoover attractor\n       aizawa              Aizawa attractor\n       three_cell_cnn      Three Cell CNN attractor\n       bouali_type_1       Bouali Type 1 attractor\n       bouali_type_2       Bouali Type 2 attractor\n       bouali_type_3       Bouali Type 3 attractor\n       finance             Finance attractor\n       burke_shaw          Burke Shaw attractor\n       moore_spiegel       Moore Spiegel attractor\n       sakarya             Sakarya attractor\n       dadras              Dadras attractor\n       halvorsen           Halvorsen attractor\n       hadley              Hadley attractor\n       chen                Chen attractor\n       chen_lee            Chen Lee attractor\n       chen_celikovsky     Chen Celikovsky attractor\n\nEach attractor also has its own parameters to set. The settings for each\nattractor can be obtained by the help command:\n``attractors ATTRACTOR -h``\n\nThe help message for Lorenz attractors will be as follows:\n\n::\n\n   usage: attractors lorenz [-h] [--sigma SIGMA] [--beta BETA] [--rho RHO]\n                            [--initcoord INITCOORD] [--xlim XLIM] [--ylim YLIM]\n                            [--zlim ZLIM]\n\n   optional arguments:\n     -h, --help            show this help message and exit\n\n   Lorenz attractor parameters:\n     --sigma SIGMA         Parameter for Lorenz attractor Default: 5\n     --beta BETA           Parameter for Lorenz attractor Default: 2.66667\n     --rho RHO             Parameter for Lorenz attractor Default: 28\n     --initcoord INITCOORD\n                           Initial coordinate for Lorenz attractor. Input format:\n                           "x,y,z" Default: [0.1, 0.1, 0.1]\n     --xlim XLIM           x axis limits for figure. Input format: "xmin,xmax"\n                           Default: [-20, 20]\n     --ylim YLIM           y axis limits for figure. Input format: "ymin,ymax"\n                           Default: [-30, 30]\n     --zlim ZLIM           z axis limits for figure. Input format: "zmin,zmax"\n                           Default: [5, 45]\n\nLicense\n-------\n\nThis package is licensed under the MIT License\n',
    'author': 'Vignesh Mohan',
    'author_email': 'vignesh.desmond@gmail.com',
    'maintainer': 'Vignesh Mohan',
    'maintainer_email': 'vignesh.desmond@gmail.com',
    'url': 'https://github.com/Vignesh-Desmond/attractors',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import matplotlib.colors as mcolors
import numpy as np


def hex_to_rgb(value):
    """
    Converts hex to normalized rgb colours

    Args:
        value (str): string of 6 characters representing a hex colour

    Returns:

    """
    value = value.strip("#")
    lv = len(value)
    rgb_vals = tuple(
        int(value[i : i + lv // 3], 16) for i in range(0, lv, lv // 3)  # noqa: E203
    )
    return [v / 256 for v in rgb_vals]


def get_continuous_cmap(hex_list, float_list=None):
    """
    Creates and returns a color map from a given hex list. If float_list
    is not provided, colour map graduates linearly between each color
    in hex_list. If float_list is provided, each color in hex_list is
    mapped to the respective location in float_list.

    Args:
        hex_list (list[str]): list of hex code strings
        float_list (list[float], optional): list of floats between 0 and 1,
            same length as hex_list. Must start with 0 and end with 1. Defaults to None.

    Returns:
        LinearSegmentedColormap: Colormap of given hex list.
    """
    rgb_list = [hex_to_rgb(i) for i in hex_list]
    if float_list:
        pass
    else:
        float_list = list(np.linspace(0, 1, len(rgb_list)))

    cdict = dict()
    for num, col in enumerate(["red", "green", "blue"]):
        col_list = [
            [float_list[i], rgb_list[i][num], rgb_list[i][num]]
            for i in range(len(float_list))
        ]
        cdict[col] = col_list
    cmp = mcolors.LinearSegmentedColormap("my_cmp", segmentdata=cdict, N=256)
    return cmp



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np

from attractors.utils.base import BaseAttractors


class DES(BaseAttractors):
    def __init__(self, initial_coord, attractor, params):
        super(DES, self).__init__(attractor, params)
        self.coord = initial_coord
        self.X = []
        self.Y = []
        self.Z = []

    def __lt__(self, other):
        if not isinstance(other, DES):
            return NotImplemented
        return len(self.X) < len(other.X)

    def __eq__(self, other):
        if not isinstance(other, DES):
            return NotImplemented
        return self.X == other.X and self.Y == other.Y and self.Z == other.Z

    def _unwrap(self, a, b, N):
        h = (b - a) / N
        timescale = np.arange(a, b, h)
        attractor_func = getattr(DES, self.attractor)
        return h, timescale, attractor_func

    def euler(self, a, b, N):
        h, ts, afunc = self._unwrap(a, b, N)

        for _ in ts:
            self.X.append(self.coord[0])
            self.Y.append(self.coord[1])
            self.Z.append(self.coord[2])

            k1 = h * afunc(self, self.coord)
            self.coord += k1

    def rk2(self, a, b, N, method):
        h, ts, afunc = self._unwrap(a, b, N)

        def heun():
            rt = self.coord
            k1 = h * afunc(self, self.coord)

            self.coord = self.coord + k1
            k2 = h * afunc(self, self.coord)
            self.coord = rt

            self.coord += (k1 + k2) / 2

        def imp_poly():
            rt = self.coord
            k1 = h * afunc(self, self.coord)

            self.coord = self.coord + k1 / 2
            k2 = h * afunc(self, self.coord)
            self.coord = rt

            self.coord += k2

        def ralston():
            rt = self.coord
            k1 = h * afunc(self, self.coord)

            self.coord = self.coord + 3 * k1 / 4
            k2 = h * afunc(self, self.coord)
            self.coord = rt

            self.coord += (k1 + 2 * k2) / 3

        for _ in ts:
            self.X.append(self.coord[0])
            self.Y.append(self.coord[1])
            self.Z.append(self.coord[2])
            eval(method)()

    def rk3(self, a, b, N):
        h, ts, afunc = self._unwrap(a, b, N)

        for _ in ts:
            self.X.append(self.coord[0])
            self.Y.append(self.coord[1])
            self.Z.append(self.coord[2])

            rt = self.coord
            k1 = h * afunc(self, self.coord)

            self.coord = self.coord + k1 / 2
            k2 = h * afunc(self, self.coord)
            self.coord = rt

            self.coord = self.coord - k1 + 2 * k2
            k3 = h * afunc(self, self.coord)
            self.coord = rt

            self.coord += (k1 + 4 * k2 + k3) / 6

    def rk4(self, a, b, N):
        h, ts, afunc = self._unwrap(a, b, N)

        for _ in ts:
            self.X.append(self.coord[0])
            self.Y.append(self.coord[1])
            self.Z.append(self.coord[2])

            rt = self.coord
            k1 = h * afunc(self, self.coord)

            self.coord = self.coord + k1 / 2
            k2 = h * afunc(self, self.coord)
            self.coord = rt

            self.coord = self.coord + k2 / 2
            k3 = h * afunc(self, self.coord)
            self.coord = rt

            self.coord = self.coord + k3
            k4 = h * afunc(self, self.coord)
            self.coord = rt

            self.coord += (k1 + 2 * k2 + 2 * k3 + k4) / 6

    def rk5(self, a, b, N):
        h, ts, afunc = self._unwrap(a, b, N)

        for _ in ts:
            self.X.append(self.coord[0])
            self.Y.append(self.coord[1])
            self.Z.append(self.coord[2])

            rt = self.coord
            k1 = h * afunc(self, self.coord)

            self.coord = self.coord + k1 / 4
            k2 = h * afunc(self, self.coord)
            self.coord = rt

            self.coord = self.coord + k2 / 8 + k1 / 8
            k3 = h * afunc(self, self.coord)
            self.coord = rt

            self.coord = self.coord + k3 - k2 / 2 + k3
            k4 = h * afunc(self, self.coord)
            self.coord = rt

            self.coord = self.coord - 3 * k1 / 16 + 9 * k4 / 16
            k5 = h * afunc(self, self.coord)
            self.coord = rt

            self.coord = (
                self.coord
                - 3 * k1 / 7
                + 2 * k2 / 7
                + 12 * k3 / 7
                - 12 * k4 / 7
                + 8 * k5 / 7
            )
            k6 = h * afunc(self, self.coord)
            self.coord = rt

            self.coord += (7 * k1 + 32 * k3 + 12 * k4 + 32 * k5 + 7 * k6) / 90



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import json
from argparse import SUPPRESS

import numpy as np

from attractors.anim.gradient import animate_gradient
from attractors.anim.sim import animate_simulation
from attractors.utils.base import ATTRACTOR_PARAMS

try:
    import importlib.metadata as metadata
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources
    import importlib_metadata as metadata

from attractors import data


def case_convert(snakecase_string):
    return snakecase_string.replace("_", " ").title().replace("Cnn", "CNN")


def cli():
    parser = argparse.ArgumentParser(add_help=False)
    required = parser.add_argument_group("required arguments")
    optional = parser.add_argument_group("other arguments")

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=metadata.version("attractors"),
    )

    parser.add_argument(
        "-h",
        "--help",
        action="help",
        default=SUPPRESS,
        help="show this help message and exit",
    )

    required.add_argument(
        "-t",
        "--type",
        help="choose simulation type",
        type=str,
        choices=["multipoint", "gradient"],
        required=True,
    )

    optional.add_argument(
        "--des",
        help="choose the Differential Equation Solver. Default: rk4",
        type=str,
        choices=["euler", "rk2", "rk3", "rk4", "rk5"],
        default="rk4",
    )

    optional.add_argument(
        "--width",
        help="set width of the figure Default: 16",
        type=int,
        default=16,
    )
    optional.add_argument(
        "--height",
        help="set height of the figure Default: 9",
        type=int,
        default=9,
    )
    optional.add_argument(
        "--dpi",
        help="set DPI of the figure Default: 120",
        type=int,
        default=120,
    )
    optional.add_argument(
        "--theme", help="choose theme (color palette) to be used", type=str
    )
    required.add_argument(
        "-s",
        "--simtime",
        help="set the simulation time",
        type=int,
        required=True,
    )
    required.add_argument(
        "-p",
        "--simpoints",
        help="set the number of points to be used for the simulation",
        type=int,
        required=True,
    )
    optional.add_argument(
        "--bgcolor",
        help=(
            "background color for figure in hex. Overrides theme settings if"
            " specified Default: #000000"
        ),
        type=str,
    )
    optional.add_argument(
        "--cmap",
        help=(
            "matplotlib cmap for palette. Overrides theme settings if"
            " specified Default: jet"
        ),
        type=str,
    )
    optional.add_argument(
        "--fps",
        help="set FPS for animated video (or interactive plot) Default: 60",
        type=int,
        default=60,
    )
    optional.add_argument(
        "--n",
        help="number of initial points for Multipoint animation Default: 3",
        type=int,
        default=3,
    )
    optional.add_argument(
        "--rk2",
        help=(
            "method for 2nd order Runge-Kutta if specified to be used." " Default: heun"
        ),
        type=str,
        default="heun",
        choices=["heun", "imp_poly", "ralston"],
    )
    optional.add_argument(
        "--outf",
        help="output video filename Default: output.mp4",
        type=str,
        default="output.mp4",
    )
    optional.add_argument(
        "--live",
        help="live plotting instead of generating video.",
        action="store_true",
    )

    subparsers = parser.add_subparsers(
        title="Attractor settings",
        description="Choose one of the attractors and specify its parameters",
        dest="attractor",
        metavar="ATTRACTOR",
    )

    for attr, attrparams in ATTRACTOR_PARAMS.items():
        attrparser = subparsers.add_parser(
            f"{attr}", help=f"{case_convert(attr)} attractor"
        )
        attrgroup = attrparser.add_argument_group(
            title=f"{case_convert(attr)} attractor parameters"
        )

        for i in range(len(attrparams["params"])):
            attrgroup.add_argument(
                f"--{attrparams['params'][i]}",
                help=(
                    f"Parameter for {case_convert(attr)} attractor "
                    f"Default: {attrparams['default_params'][i]}"
                ),
                type=float,
                default=attrparams["default_params"][i],
            )
        attrgroup.add_argument(
            "--initcoord",
            help=(
                f"Initial coordinate for {case_convert(attr)} attractor. Input"
                f" format: \"x y z\" Default: {attrparams['init_coord']}"
            ),
            type=float,
            nargs=3,
            default=attrparams["init_coord"],
        )
        for k in ["x", "y", "z"]:
            attrgroup.add_argument(
                f"--{k}lim",
                help=(
                    f"{k} axis limits for figure. Input format:"
                    f" \"{k}min {k}max\" Default: {attrparams[f'{k}lim']}"
                ),
                type=float,
                nargs=2,
                default=attrparams[f"{k}lim"],
            )

    args = parser.parse_args()

    # * load theme
    raw_themes_data = pkg_resources.open_text(data, "themes.json")
    themes = json.load(raw_themes_data)

    # * load args
    attractor = args.attractor
    theme = themes[args.theme] if args.theme is not None else None
    bgcolor = "#000000"
    palette = "jet"
    if theme is not None:
        palette_temp = list(theme.values())
        palette_temp.remove(theme["background"])
        bgcolor = args.bgcolor if args.bgcolor is not None else theme["background"]
        palette = args.cmap if args.cmap is not None else palette_temp
    width = args.width
    height = args.height
    dpi = args.dpi
    fps = args.fps
    simtime = args.simtime
    simpoints = args.simpoints
    n = args.n
    des = args.des
    rk2_method = args.rk2
    outf = args.outf
    live = args.live

    attr = ATTRACTOR_PARAMS[attractor]
    init_coord = np.array(args.initcoord, dtype="double")
    attr_params = {p: getattr(args, p) for p in attr["params"]}
    xlim = args.xlim
    ylim = args.ylim
    zlim = args.zlim

    if args.type == "multipoint":
        animate_simulation(
            attractor,
            init_coord,
            attr_params,
            xlim,
            ylim,
            zlim,
            width,
            height,
            dpi,
            bgcolor,
            palette,
            simtime,
            simpoints,
            n,
            des,
            live,
            rk2_method,
            fps,
            outf,
        )
    else:
        animate_gradient(
            attractor,
            init_coord,
            attr_params,
            xlim,
            ylim,
            zlim,
            width,
            height,
            dpi,
            bgcolor,
            palette,
            simtime,
            simpoints,
            des,
            rk2_method,
            fps,
            outf,
        )


if __name__ == " __main__":
    cli()
